/***************************************************************************
 *   Copyright (C) 2003 by Sbastien Laot                                 *
 *   slaout@linux62.org                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qpainter.h>
#include <kglobalsettings.h>
#include <qstyle.h>
#include <kapplication.h>
#include <kstyle.h>
#include <qcursor.h>
#include <kiconloader.h>
#include <kpixmapeffect.h>
#include <kpixmap.h>
#include <kglobal.h>
#include <klocale.h>
#include <kurifilter.h>
#include <qfile.h>

#include <stdlib.h> // rand() function

#include <iostream>

#include "basket.h"
#include "tag.h"
#include "item.h"
#include "utilities.h"
#include "settings.h"

/* Helping function: */

bool tooDark(const QColor &color)
{
	int dontCare, value;
	color.getHsv(/*hue:*/dontCare, /*saturation:*/dontCare, value);
	return value < 175;
}

/** class Item: */

int Item::ITEM_MARGIN      = 2;
int Item::INSERTION_HEIGHT = 5;
int Item::EXPANDER_WIDTH   = 9;
int Item::EXPANDER_HEIGHT  = 9;
int Item::GROUP_WIDTH      = 2*ITEM_MARGIN + EXPANDER_WIDTH;
int Item::HANDLE_WIDTH     = GROUP_WIDTH;
int Item::RESIZER_WIDTH    = GROUP_WIDTH;
int Item::TAG_ARROW_WIDTH  = 5;
int Item::EMBLEM_SIZE      = 16;
int Item::MIN_HEIGHT       = 2*ITEM_MARGIN + EMBLEM_SIZE;


QString Item::addedStringDate()
{
	return KGlobal::locale()->formatDateTime(m_addedDate);
}

QString Item::lastModificationStringDate()
{
	return KGlobal::locale()->formatDateTime(m_lastModificationDate);
}


QString Item::toText(const QString &cuttedFullPath)
{
	if (content()) {
		// Convert item to text:
		QString text = content()->toText(cuttedFullPath);
		// If we should not export tags with the text, return immediatly:
		if (!Settings::exportTextTags())
			return text;
		// Compute the text equivalent of the tag states:
		QString firstLine;
		QString otherLines;
		for (State::List::Iterator it = m_states.begin(); it != m_states.end(); ++it) {
			if (!(*it)->textEquivalent().isEmpty()) {
				firstLine += (*it)->textEquivalent() + " ";
				if ((*it)->onAllTextLines())
					otherLines += (*it)->textEquivalent() + " ";
			}
		}
		// Merge the texts:
		if (firstLine.isEmpty())
			return text;
		if (otherLines.isEmpty())
			return firstLine + text;
		QStringList lines = QStringList::split('\n', text, /*allowEmptyEntries=*/true);
		QString result = firstLine + lines[0] + (lines.count() > 1 ? "\n" : "");
		for (uint i = 1/*Skip the first line*/; i < lines.count(); ++i)
			result += otherLines + lines[i] + (i < lines.count() - 1 ? "\n" : "");
		return result;
	} else
		return "";
}



Item::Item(Basket *parent)
 : m_prev(0), m_next(0),
   m_x(0), m_y(-1), m_width(-1), m_height(-1),
   m_groupWidth(250),
   m_isFolded(false), m_firstChild(0L), m_parentItem(0),
   m_basket(parent), m_content(0), m_addedDate(QDateTime::currentDateTime()), m_lastModificationDate(QDateTime::currentDateTime()),
   m_computedAreas(false), m_onTop(false),
   m_deltaX(0), m_deltaY(0), m_deltaHeight(0), m_collapseFinished(true),
   m_hovered(false), m_focused(false), m_selected(false), m_wasInLastSelectionRect(false),
   m_computedState(), m_emblemsCount(0), m_haveInvisibleTags(false),
   m_matching(true)
{
}

Item::~Item()
{
}

bool Item::computeMatching(const SearchData &data)
{
	// Groups are always matching:
	if (!content())
		return true;

	bool matching;
	// First match tags (they are fast to compute):
	switch (data.tagSearchType) {
		default:
			case SearchData::DontCareTagsSearch: matching = true;                   break;
			case SearchData::NotTaggedSearch:    matching = m_states.count() <= 0;  break;
			case SearchData::TaggedSearch:       matching = m_states.count() > 0;   break;
			case SearchData::TagSearch:          matching = hasTag(data.tag);       break;
			case SearchData::StateSearch:        matching = hasState(data.state);   break;
	}

	// Don't try to match the content text if we are not matching now (the search is of 'AND' type) or if we shouldn't try to match the string:
	if (matching && !data.string.isEmpty())
		matching = content()->match(data);

	return matching;
}

int Item::newSearch(const SearchData &data)
{
	bool wasMatching = matching();
	m_matching = computeMatching(data);
	setOnTop(wasMatching && matching());
	if (!matching())
		setSelected(false);

	int countMatches = (content() && matching() ? 1 : 0);

	Item *child = firstChild();
	while (child) {
		countMatches += child->newSearch(data);
		child = child->next();
	}

	return countMatches;
}

void Item::deleteSelectedItems(bool deleteFilesToo)
{
	if (content() && isSelected()) {
		basket()->unplugItem(this);
		if (deleteFilesToo && content() && content()->useFile())
			//QFile::remove(fullPath());
			Utilities::deleteRecursively(fullPath());//basket()->deleteFiles(fullPath()); // Also delete the folder if it's a folder
		//delete this;
		return;
	}

	Item *child = firstChild();
	Item *next;
	while (child) {
		next = child->next(); // If we delete 'child' on the next line, child->next() will be 0!
		child->deleteSelectedItems(deleteFilesToo);
		child = next;
	}
}

int Item::count()
{
	if (content())
		return 1;

	int count = 0;
	for (Item *child = firstChild(); child; child = child->next())
		count += child->count();
	return count;
}

// TODO:

#define FOR_EACH_CHILD(childVar) \
	for (Item *childVar = firstChild(); childVar; childVar = childVar->next())

#define FOR_EACH_VISIBLE_CHILD(childVar) \
	for (...)

#define FOR_EACH_CHILD_DO \
	for (Item *childVar = firstChild(); childVar; childVar = childVar->next()) \
		childVar

int Item::countDirectChilds()
{
	int count = 0;
	FOR_EACH_CHILD (child)
		++count;
	return count;
}

QString Item::fullPath()
{
	if (content())
		return basket()->fullPath() + content()->fileName();
	else
		return "";
}

void Item::update()
{
	basket()->updateItem(this);
}

/*void Item::doSomething()
{
	m_matching = true;

	FOR_EACH_CHILD (child)
		child->doSomething();

	FOR_EACH_CHILD_DO->doSomething();
}*/


void Item::setFocused(bool focused)
{
	if (m_focused == focused)
		return;

	m_focused = focused;
	unbufferize();
	basket()->updateItem(this); // FIXME: ???????
}

void Item::setSelected(bool selected)
{
	if (isGroup())
		selected = false; // A group cannot be selected!

	if (m_selected == selected)
		return;

//	if (!isGroup()) {
		if (selected)
			basket()->addSelectedItem();
		else
			basket()->removeSelectedItem();
//	}

	m_selected = selected;
	unbufferize();
	basket()->updateItem(this); // FIXME: ???????
}

void Item::resetWasInLastSelectionRect()
{
	m_wasInLastSelectionRect = false;

	Item *child = firstChild();
	while (child) {
		child->resetWasInLastSelectionRect();
		child = child->next();
	}
}

void Item::selectIn(const QRect &rect, bool invertSelection)
{
	QRect myRect(x(), y(), width(), height());
	bool intersects = myRect.intersects(rect);
	bool toSelect = intersects;
	if (invertSelection) {
		if (m_wasInLastSelectionRect == intersects)
			toSelect = isSelected();
		else if (intersects xor m_wasInLastSelectionRect)
			toSelect = !isSelected();// xor intersects;
	}
	setSelected(toSelect);
	m_wasInLastSelectionRect = intersects;

	Item *child = firstChild();
	bool first = true;
	while (child) {
		if ((showSubItems() || first) && child->matching())
			child->selectIn(rect, invertSelection);
		else
			child->setSelectedRecursivly(false);
		child = child->next();
		first = false;
	}
}

bool Item::allSelected()
{
	if (isGroup()) {
		Item *child = firstChild();
		bool first = true;
		while (child) {
			if ((showSubItems() || first) && child->matching())
				if (!child->allSelected())
					return false;;
			child = child->next();
			first = false;
		}
		return true;
	} else
		return isSelected();
}

void Item::setSelectedRecursivly(bool selected)
{
	setSelected(selected && matching());

	Item *child = firstChild();
	while (child) {
		child->setSelectedRecursivly(selected);
		child = child->next();
	}
}

void Item::invertSelectionRecursivly()
{
	if (content())
		setSelected(!isSelected() && matching());

	Item *child = firstChild();
	while (child) {
		child->invertSelectionRecursivly();
		child = child->next();
	}
}

void Item::unselectAllBut(Item *toSelect)
{
	if (this == toSelect)
		setSelectedRecursivly(true);
	else {
		setSelected(false);

		Item *child = firstChild();
		bool first = true;
		while (child) {
			if ((showSubItems() || first) && child->matching())
				child->unselectAllBut(toSelect);
			else
				child->setSelectedRecursivly(false);
			child = child->next();
			first = false;
		}
	}
}

void Item::invertSelectionOf(Item *toSelect)
{
	if (this == toSelect)
		setSelectedRecursivly(!isSelected());
	else {
		Item *child = firstChild();
		bool first = true;
		while (child) {
			if ((showSubItems() || first) && child->matching())
				child->invertSelectionOf(toSelect);
			child = child->next();
			first = false;
		}
	}
}

Item* Item::theSelectedItem()
{
	if (!isGroup() && isSelected())
		return this;

	Item *selectedOne;
	Item *child = firstChild();
	while (child) {
		selectedOne = child->theSelectedItem();
		if (selectedOne)
			return selectedOne;
		child = child->next();
	}

	return 0;
}

ItemSelection* Item::selectedItems()
{
	if (content())
		if (isSelected())
			return new ItemSelection(this);
		else
			return 0;

	ItemSelection *selection = new ItemSelection(this);

	FOR_EACH_CHILD (child)
		selection->append(child->selectedItems());

	if (selection->firstChild) {
		if (selection->firstChild->next)
			return selection;
		else {
			// If 'selection' is a groupe with only one content, return directly that content:
			ItemSelection *reducedSelection = selection->firstChild;
// 			delete selection; // TODO: Cut all connexions of 'selection' before deleting it!
			for (ItemSelection *node = reducedSelection; node; node = node->next)
				node->parent = 0;
			return reducedSelection;
		}
	} else {
		delete selection;
		return 0;
	}
}

Item* Item::firstRealChild()
{
	Item *child = m_firstChild;
	while (child) {
		if ( !child->isGroup() /*&& child->matching()*/ )
			return child;
		child = child->firstChild();
	}
	// Empty group:
	return NULL;
}

Item* Item::lastChild()
{
	Item *child = m_firstChild;
	while (child && child->next())
		child = child->next();

	return child;
}

Item* Item::lastSibling()
{
	Item *last = this;
	while (last && last->next())
		last = last->next();

	return last;
}

int Item::yExpander()
{
	Item *child = firstRealChild();
	if (child)
		return (child->height() - EXPANDER_HEIGHT) / 2 + !(child->height()%2);
	else // Groups always have at least 2 items, except for columns which can have no child (but should exists anyway):
		return 0;
}

bool Item::isFree()
{
	return parentItem() == 0 && basket()->isFreeLayout();
}

bool Item::isColumn()
{
	return parentItem() == 0 && basket()->isColumnsLayout();
}

bool Item::hasResizer()
{
	// "isFree" || "isColmun but not the last"
	return parentItem() == 0 && (basket()->isFreeLayout() || m_next != 0L);
}

int Item::resizerHeight()
{
	return (isColumn() ? basket()->contentsHeight() : height());
}

void Item::setHoveredZone(Zone zone) // TODO: Remove setHovered(bool) and assume it is hovered if zone != None !!!!!!!
{
	if (m_hoveredZone != zone) {
		if (content())
			content()->setHoveredZone(m_hoveredZone, zone);
		m_hoveredZone = zone;
		unbufferize();
	}
}

Item::Zone Item::zoneAt(const QPoint &pos, bool toAdd)
{
	// Keep the resizer highlighted when resizong, even if the cursor is over another item:
	if (basket()->resizingItem() == this)
		return Resizer;

	// When dropping/pasting something on a column resizer, add it at the bottom of the column, and don't group it whith the whole column:
	if (toAdd && isColumn() && hasResizer()) {
		int right = rightLimit() - x();
		if ((pos.x() >= right) && (pos.x() < right + RESIZER_WIDTH) && (pos.y() >= 0) && (pos.y() < resizerHeight())) // Code copied from below
			return BottomColumn;
	}

	// Below a column:
	if (isColumn()) {
		if (pos.y() >= height() && pos.x() < rightLimit() - x())
			return BottomColumn;
	}

	// If toAdd, return only TopInsert, TopGroup, BottomInsert or BottomGroup
	// (by spanning those areas in 4 equal rectangles in the item):
	if (toAdd) {
		if (isColumn() && pos.y() >= height())
			return BottomGroup;
		if (pos.y() < height() / 2)
			if (pos.x() < width() / 2 && !isFree())
				return TopInsert;
			else
				return TopGroup;
		else
			if (pos.x() < width() / 2 && !isFree())
				return BottomInsert;
			else
				return BottomGroup;
	}

	// If in the resizer:
	if (hasResizer()) {
		int right = rightLimit() - x();
		if ((pos.x() >= right) && (pos.x() < right + RESIZER_WIDTH) && (pos.y() >= 0) && (pos.y() < resizerHeight()))
			return Resizer;
	}

	// If isGroup, return only Group, GroupExpander, TopInsert or BottomInsert:
	if (isGroup()) {
		if (pos.y() < INSERTION_HEIGHT)
			return (isFree() ? TopGroup : TopInsert);
		if (pos.y() >= height() - INSERTION_HEIGHT)
			return (isFree() ? BottomGroup : BottomInsert);

		if (pos.x() >= ITEM_MARGIN  &&  pos.x() < ITEM_MARGIN + EXPANDER_WIDTH) {
			int yExp = yExpander();
			if (pos.y() >= yExp  &&  pos.y() < yExp + EXPANDER_HEIGHT)
				return GroupExpander;
		}
		if (pos.x() < width())
			return Group;
		else
			return None;
	}

	// Else, it's a normal item:

	if (pos.x() < HANDLE_WIDTH)
		return Handle;

	if (pos.y() < INSERTION_HEIGHT)
		if (pos.x() < width() / 2 && !isFree())
			return TopInsert;
		else
			return TopGroup;

	if (pos.y() >= height() - INSERTION_HEIGHT)
		if (pos.x() < width() / 2 && !isFree())
			return BottomInsert;
		else
			return BottomGroup;

	for (int i =0; i < m_emblemsCount; i++) {
		if ( pos.x() >= HANDLE_WIDTH + (ITEM_MARGIN+EMBLEM_SIZE)*i  &&
				   pos.x() <  HANDLE_WIDTH + (ITEM_MARGIN+EMBLEM_SIZE)*i + ITEM_MARGIN+EMBLEM_SIZE )
			return (Zone)(Emblem0 + i);
	}

	if (pos.x() < HANDLE_WIDTH + (ITEM_MARGIN+EMBLEM_SIZE)*m_emblemsCount + ITEM_MARGIN + TAG_ARROW_WIDTH + ITEM_MARGIN)
		return TagsArrow;

	if (!linkAt(pos).isEmpty())
		return Link;

	int customZone = content()->zoneAt(pos - QPoint(contentX(), ITEM_MARGIN));
	if (customZone)
		return (Item::Zone)customZone;

	return Content;
}

QString Item::linkAt(const QPoint &pos)
{
	QString link = m_content->linkAt(pos - QPoint(contentX(), ITEM_MARGIN));
	if (link.isEmpty())
		return link;
	else
		return KURIFilter::self()->filteredURI(link);
}

int Item::contentX()
{
	return HANDLE_WIDTH + ITEM_MARGIN + (EMBLEM_SIZE+ITEM_MARGIN)*m_emblemsCount + TAG_ARROW_WIDTH + ITEM_MARGIN;
}

QRect Item::zoneRect(Item::Zone zone, const QPoint &pos)
{
	if (zone >= Emblem0)
		return QRect(HANDLE_WIDTH + (ITEM_MARGIN+EMBLEM_SIZE)*(zone-Emblem0),
					 INSERTION_HEIGHT,
					 ITEM_MARGIN + EMBLEM_SIZE,
					 height() - 2*INSERTION_HEIGHT);

	int yExp;
	int right;
	int xGroup = (isFree() ? (isGroup() ? 0 : GROUP_WIDTH) : width() / 2);
	QRect rect;
	switch (zone) {
		case Item::Handle:        return QRect(0, 0, HANDLE_WIDTH, height());
		case Item::Group:
			yExp = yExpander();
			if (pos.y() < yExp)                   return QRect(0,                     INSERTION_HEIGHT,       width(),     yExp - INSERTION_HEIGHT);
			if (pos.y() > yExp + EXPANDER_HEIGHT) return QRect(0,                     yExp + EXPANDER_HEIGHT, width(),     height() - yExp - EXPANDER_HEIGHT - INSERTION_HEIGHT);
			if (pos.x() < ITEM_MARGIN)            return QRect(0,                     0,                      ITEM_MARGIN, height());
			else                                  return QRect(width() - ITEM_MARGIN, 0,                      ITEM_MARGIN, height());
		case Item::TagsArrow:     return QRect(HANDLE_WIDTH + (ITEM_MARGIN+EMBLEM_SIZE)*m_emblemsCount,
		                                       INSERTION_HEIGHT,
		                                       ITEM_MARGIN + TAG_ARROW_WIDTH + ITEM_MARGIN,
		                                       height() - 2*INSERTION_HEIGHT);
		case Item::Custom0:
		case Item::Content:       rect = content()->zoneRect(zone, pos - QPoint(contentX(), ITEM_MARGIN));
		                          rect.moveBy(contentX(), ITEM_MARGIN);
		                          return rect.intersect(  QRect(contentX(), INSERTION_HEIGHT, width() - contentX(), height() - 2*INSERTION_HEIGHT)  ); // Only IN contentRect
		case Item::GroupExpander: return QRect(ITEM_MARGIN, yExpander(), EXPANDER_WIDTH, EXPANDER_HEIGHT);
		case Item::Resizer:       right = rightLimit();
		                          return QRect(right - x(), 0, RESIZER_WIDTH, resizerHeight());
		case Item::Link:
//		case Item::Content:       return QRect(contentX(), INSERTION_HEIGHT,
//		                                       width() - contentX(), height() - 2*INSERTION_HEIGHT);
		case Item::TopInsert:     if (isGroup()) return QRect(0,            0,                           width(),                    INSERTION_HEIGHT);
		                          else           return QRect(HANDLE_WIDTH, 0,                           width() / 2 - HANDLE_WIDTH, INSERTION_HEIGHT);
		case Item::TopGroup:    /*if (isGroup()) return QRect(width() / 2,  0,                           width() / 2,                INSERTION_HEIGHT);
		                          else*/         return QRect(xGroup,       0,                           width() - xGroup,           INSERTION_HEIGHT);
		case Item::BottomInsert:  if (isGroup()) return QRect(0,            height() - INSERTION_HEIGHT, width(),                    INSERTION_HEIGHT);
		                          else           return QRect(HANDLE_WIDTH, height() - INSERTION_HEIGHT, width() / 2 - HANDLE_WIDTH, INSERTION_HEIGHT);
		case Item::BottomGroup: /*if (isGroup()) return QRect(width() / 2,  height() - INSERTION_HEIGHT, width() / 2,                INSERTION_HEIGHT);
		                          else*/         return QRect(xGroup,       height() - INSERTION_HEIGHT, width() - xGroup,           INSERTION_HEIGHT);
		case Item::BottomColumn:  return QRect(0, height(), rightLimit() - x(), basket()->contentsHeight() - height());
		case Item::None:          return QRect(/*0, 0, -1, -1*/);
		default:                  return QRect(/*0, 0, -1, -1*/);
	}
}

void Item::setCursor(Zone zone)
{
	switch (zone) {
		case Item::Handle:
		case Item::Group:         basket()->viewport()->setCursor(Qt::SizeAllCursor);      break;
		case Item::Resizer:       if (isColumn())
		                              basket()->viewport()->setCursor(Qt::SplitHCursor);
		                          else
		                              basket()->viewport()->setCursor(Qt::SizeHorCursor);  break;

		case Item::Custom0:       content()->setCursor(basket()->viewport(), zone);        break;

		case Item::Link:
		case Item::TagsArrow:
		case Item::GroupExpander: basket()->viewport()->setCursor(Qt::PointingHandCursor); break;

		case Item::Content:       basket()->viewport()->setCursor(Qt::IbeamCursor);        break;

		case Item::TopInsert:
		case Item::TopGroup:
		case Item::BottomInsert:
		case Item::BottomGroup:
		case Item::BottomColumn:  basket()->viewport()->setCursor(Qt::CrossCursor);        break;
		case Item::None:          basket()->viewport()->unsetCursor(); break;
		default:
			State *state = stateForEmblemNumber(zone - Emblem0);
			if (state && state->parentTag()->states().count() > 1)
				basket()->viewport()->setCursor(Qt::PointingHandCursor);
			else
				basket()->viewport()->unsetCursor();
	}
}

void Item::addAnimation(int deltaX, int deltaY, int deltaHeight)
{
	// Don't process animation that make the item stay in place!
	if (deltaX == 0 && deltaY == 0 && deltaHeight == 0)
		return;

	// If it was not animated previsouly, make it animated:
	if (m_deltaX == 0 && m_deltaY == 0 && m_deltaHeight == 0)
		basket()->addAnimatedItem(this);

	// Configure the animation:
	m_deltaX      += deltaX;
	m_deltaY      += deltaY;
	m_deltaHeight += deltaHeight;
}

void Item::setFinalPosition(int x, int y)
{
	addAnimation(x - finalX(), y - finalY());
}

void Item::initAnimationLoad()
{
	int x, y;
	switch (rand() % 4) {
		case 0: // Put it on top:
			x = basket()->contentsX() + rand() % basket()->contentsWidth();
			y = -height();
			break;
		case 1: // Put it on bottom:
			x = basket()->contentsX() + rand() % basket()->contentsWidth();
			y = basket()->contentsY() + basket()->visibleHeight();
			break;
		case 2: // Put it on left:
			x = -width() - (hasResizer() ? Item::RESIZER_WIDTH : 0);
			y = basket()->contentsY() + rand() % basket()->visibleHeight();
			break;
		case 3: // Put it on right:
		default: // In the case of...
			x = basket()->contentsX() + basket()->visibleWidth();
			y = basket()->contentsY() + rand() % basket()->visibleHeight();
			break;
	}
	cancelAnimation();
	addAnimation(finalX() - x, finalY() - y);
	setX(x);
	setY(y);

	if (isGroup()) {
		const int viewHeight = basket()->contentsY() + basket()->visibleHeight();
		Item *child = firstChild();
		bool first = true;
		while (child) {
			if (child->finalY() < viewHeight) {
				if ((showSubItems() || first) && child->matching())
					child->initAnimationLoad();
			} else
				break; // 'child' are not a free items (because child of at least one item, 'this'), so 'child' is ordered vertically.
			child = child->next();
			first = false;
		}
	}
}


bool Item::advance()
{
	// Animate X:
	if (m_deltaX != 0) {
		int deltaX = m_deltaX / 3;
		if (deltaX == 0)
			deltaX = (m_deltaX > 0 ? 1 : -1);
		setX(m_x + deltaX);
		m_deltaX -= deltaX;
	}

	// Animate Y:
	if (m_deltaY != 0) {
		int deltaY = m_deltaY / 3;
		if (deltaY == 0)
			deltaY = (m_deltaY > 0 ? 1 : -1);
		setY(m_y + deltaY);
		m_deltaY -= deltaY;
	}

	// Animate Height:
	if (m_deltaHeight != 0) {
		int deltaHeight = m_deltaHeight / 3;
		if (deltaHeight == 0)
			deltaHeight = (m_deltaHeight > 0 ? 1 : -1);
		m_height += deltaHeight;
		unbufferize();
		m_deltaHeight -= deltaHeight;
	}

	if (m_deltaHeight == 0)
		m_collapseFinished = true;

	// Return true if the animation is finished:
	return (m_deltaX == 0 && m_deltaY == 0 && m_deltaHeight == 0);
}

void Item::unsetWidth()
{
	m_width = 0;
	unbufferize();

	Item *child = firstChild();
	while (child) {
		child->unsetWidth();
		child = child->next();
	}
}

void Item::requestRelayout()
{
	m_width = 0;
	unbufferize();
	basket()->relayoutItems(true); // TODO: A signal that will relayout ONCE and DELAYED if called several times
}

void Item::setWidth(int width) // TODO: inline ?
{
	if (m_width != width)
		setWidthForceRelayout(width);
}

void Item::setWidthForceRelayout(int width)
{
	unbufferize();
	m_width = (width < minWidth() ? minWidth() : width);
	int contentWidth = width - contentX() - ITEM_MARGIN;
	if (m_content) { ///// FIXME: is this OK?
		m_height = m_content->setWidthAndGetHeight(contentWidth < 1 ? 1 : contentWidth) + 2 * ITEM_MARGIN;
		if (m_height < 3 * INSERTION_HEIGHT) // Assure a minimal size...
			m_height = 3 * INSERTION_HEIGHT;
	}
}

int Item::minWidth()
{
	if (m_content)
		return contentX() + m_content->minWidth()/*m_minContentWidth*/ + ITEM_MARGIN;
	else
		return GROUP_WIDTH; ///// FIXME: is this OK?
}

int Item::minRight()
{
	if (isGroup()) {
		int right = finalX() + width();
		Item* child = firstChild();
		bool first = true;
		while (child) {
			if ((showSubItems() || first) && child->matching())
				right = QMAX(right, child->minRight());
			child = child->next();
			first = false;
		}
		if (isColumn()) {
			int minColumnRight = finalX() + 2*HANDLE_WIDTH;
			if (right < minColumnRight)
				return minColumnRight;
		}
		return right;
	} else
		return finalX() + minWidth();
}

void Item::setX(int x)
{
	if (m_x == x)
		return;

	if (isBufferized() && basket()->hasBackgroundImage()) {
		// Unbufferize only if the background change:
		if (basket()->isTiledBackground())
			unbufferize();
		else {
			int bgw = basket()->backgroundPixmap()->width();
			if (m_x >= bgw && x < bgw) // Was not in the background image and is now inside it:
				unbufferize();
			else if (m_x < bgw) // Was in the background image and is now at another position of the background image or is now outside:
				unbufferize();
		}
	}

	m_x = x;
}

void Item::setY(int y)
{
	if (m_y == y)
		return;

	if (isBufferized() && basket()->hasBackgroundImage()) {
		// Unbufferize only if the background change:
		if (basket()->isTiledBackground())
			unbufferize();
		else {
			int bgh = basket()->backgroundPixmap()->height();
			if (m_y >= bgh && y < bgh) // Was not in the background image and is now inside it:
				unbufferize();
			else if (m_y < bgh) // Was in the background image and is now at another position of the background image or is now outside:
				unbufferize();
		}
	}

	m_y = y;
}


void Item::toggleFolded(bool animate)
{
	m_isFolded = ! m_isFolded;

	unbufferize();

	if (animate) {
		// We animate collapsing (so sub-items fluidly go under the first item)
		// We don't animate expanding: we place sub-items directly under the first item (and the next relayout will animate the expanding)
		// But if user quickly collapsed and then expand (while the collapsing animation isn't finished), we animate anyway
		bool animateSetUnder = (m_isFolded || !m_collapseFinished);
//		std::cout << "fold:" << m_isFolded << " collapseFinished:"  << m_collapseFinished << " animateSetUnder:" << animateSetUnder << std::endl;

		if (m_isFolded)
			m_collapseFinished = false;

		Item* item = firstChild();
		if (item) {
			item->setOnTop(true);
			while ( (item = item->next()) ) { // Don't process the first child: it is OK
				item->setRecursivelyUnder(firstChild(), animateSetUnder);
				item->setOnTop(false);
			}
		}
	}

	basket()->save(); // FIXME: SHOULD WE ALWAYS SAVE ????????,
}

void Item::setRecursivelyUnder(Item *under, bool animate)
{
	int y = finalHeight() > under->finalHeight() ? under->finalY() : under->finalBottom() - finalHeight() + 1;
	if (animate)
		//addAnimation(0, y - finalY()/*m_y*/);
		setFinalPosition(finalX(), y);
	else {
		setY(y);
		cancelAnimation();
	}

	if (isGroup()) {
		Item* item = firstChild();
		while (item) {
			item->setRecursivelyUnder(under, animate);
			item = item->next();
		}
	}
}


Item* Item::itemAt(int x, int y)
{
//	if (basket()->draggedItems().count() > 0 && allSelected())
//		return 0;

	if (matching() && hasResizer()) {
		int right = rightLimit();
		// TODO: This code is dupliacted 3 times: !!!!
		if ((x >= right) && (x < right + RESIZER_WIDTH) && (y >= m_y) && (y < m_y + resizerHeight())) {
//			if (!basket()->draggedItems().contains(this)) {
				if ( ! m_computedAreas )
					recomputeAreas();
				for (QValueList<QRect>::iterator it = m_areas.begin(); it != m_areas.end(); ++it) {
					QRect &rect = *it;
					if (rect.contains(x, y))
//						if (basket()->draggedItems().contains(this))
//							return 0;
//						else
							return this;
				}
//			}
		}
	}

	if (isGroup()) {
		if ((x >= m_x) && (x < m_x + width()) && (y >= m_y) && (y < m_y + m_height)) {
//			if (basket()->draggedItems().contains(this))
//				return 0;
			if ( ! m_computedAreas )
				recomputeAreas();
			for (QValueList<QRect>::iterator it = m_areas.begin(); it != m_areas.end(); ++it) {
				QRect &rect = *it;
				if (rect.contains(x, y))
					return this;
			}
			return NULL;
		}
		Item *child = firstChild();
		Item *found;
		bool first = true;
		while (child) {
			if ((showSubItems() || first) && child->matching()) {
				found = child->itemAt(x, y);
				if (found)
					return found;
			}
			child = child->next();
			first = false;
		}
	} else if (matching() && y >= m_y && y < m_y + m_height && x >= m_x && x < m_x + m_width) {
//		if (basket()->draggedItems().contains(this))
//			return 0;
		if ( ! m_computedAreas )
			recomputeAreas();
		for (QValueList<QRect>::iterator it = m_areas.begin(); it != m_areas.end(); ++it) {
			QRect &rect = *it;
			if (rect.contains(x, y))
				return this;
		}
		return NULL;
	}

	return NULL;
}

QRect Item::rect()
{
	return QRect(x(), y(), width(), height());
}

QRect Item::resizerRect()
{
	return QRect(rightLimit(), y(), RESIZER_WIDTH, resizerHeight());
}


void Item::relayoutAt(int x, int y, bool animate)
{
	if (!matching())
		return;

	m_computedAreas = false;
	m_areas.clear();

	// Don't relayout free items one under the other, because by definition they are freely positionned!
	if (isFree()) {
		x = finalX();
		y = finalY();
//		cancelAnimation();
	// If it's a column, it always have the same "fixed" position (no animation):
	} else if (isColumn()) {
		x = (prev() ? prev()->rightLimit() + RESIZER_WIDTH : 0);
		y = 0;
		cancelAnimation();
		setX(x);
		setY(y);
	// But relayout others vertically if they are inside such primary groups or if it is a "normal" basket:
	} else {
//		cancelAnimation();
		if (animate)
//			addAnimation(x - m_x, y - m_y);
			setFinalPosition(x, y);
		else {
			cancelAnimation();
			setX(x);
			setY(y);
		}
	}

	// Then, relayout sub-items (only the first, if the group is folded) and so, assign an height to the group:
	if (isGroup()) {
//		m_width = (isColumn() ? 0 : GROUP_WIDTH);
		int h = 0;
		Item *child = firstChild();
		bool first = true;
		while (child) {
			if (child->matching() && (!m_isFolded || first)) { // Don't use showSubItems() but use !m_isFolded because we don't want a relayout for the animated collapsing items
				child->relayoutAt(x + width(), y+h, animate);
				h += child->finalHeight();
			} else                                  // In case the user collapse a group, then move it and then expand it:
				child->setXRecursivly(x + width()); //  items SHOULD have a good X coordonate, and not the old one!
			if (!child->matching()) {
				// For future animation when re-match, but on bottom of already matched items!
				// Find parent primary item and set the Y to THAT y:
				Item *parentPrimaryItem = child;
				while (parentPrimaryItem->parentItem())
					parentPrimaryItem = parentPrimaryItem->parentItem();
				child->setY(parentPrimaryItem->y());
			}
			child = child->next();
			first = false;
		}
		if (finalHeight()/*m_height*/ != h || m_height != h) {
			unbufferize();
			if (animate)
				addAnimation(0, 0, h - finalHeight()/*m_height*/);
			else {
				m_height = h;
				unbufferize();
			}
		}
	} else {
		setWidth(finalRightLimit() - x);
		// If rightLimit is excedded, set the top-level right limit!!!
		// and NEED RELAYOUT
	}

	// Set the basket area limits (but not for child items: no need, because they will look for theire parent item):
	if (!parentItem()) {
		if (basket()->tmpWidth < finalRightLimit() + (hasResizer() ? RESIZER_WIDTH : 0))
			basket()->tmpWidth = finalRightLimit() + (hasResizer() ? RESIZER_WIDTH : 0);
		if (basket()->tmpHeight < finalY() + finalHeight())
			basket()->tmpHeight = finalY() + finalHeight();
	// However, if the item exceed the allowed size, let it! :
	} else if (!isGroup()) {
		if (basket()->tmpWidth < finalX() + width())
			basket()->tmpWidth = finalX() + width();
		if (basket()->tmpHeight < finalY() + finalHeight())
			basket()->tmpHeight = finalY() + finalHeight();
	}
}

void Item::setXRecursivly(int x)
{
	m_deltaX = 0;
	setX(x);

	Item *child = firstChild();
	while (child) {
		child->setXRecursivly(x + width());
		child = child->next();
	}
}

void Item::setYRecursivly(int y)
{
	m_deltaY = 0;
	setY(y);

	Item *child = firstChild();
	while (child) {
		child->setYRecursivly(y);
		child = child->next();
	}
}

void Item::setGroupWidth(int width)
{
	m_groupWidth = width;
}

int Item::groupWidth()
{
	if (hasResizer())
		return m_groupWidth;
	else
		return rightLimit() - x();
}

int Item::rightLimit()
{
	if (isColumn() && m_next == 0L) // The last column
		return QMAX(x() + minWidth(), basket()->visibleWidth());
	else if (parentItem())
		return parentItem()->rightLimit();
	else
		return m_x + m_groupWidth;
}

int Item::finalRightLimit()
{
	if (isColumn() && m_next == 0L) // The last column
		return QMAX(finalX() + minWidth(), basket()->visibleWidth());
	else if (parentItem())
		return parentItem()->finalRightLimit();
	else
		return finalX() + m_groupWidth;
}

/*void Item::prepend(Item *item)
{
	item->m_parentItem = this;
	item->setNext(m_firstChild);
	item->setPrev(0L);

	if (m_firstChild)
		m_firstChild->setPrev(item);

	m_firstChild = item;
}*/

/*
 * This code is derivated from drawMetalGradient() from the Qt documentation:
 */
void drawGradient( QPainter *p, const QColor &colorTop, const QColor & colorBottom,
				   int x, int y, int w, int h,
				   bool sunken, bool horz, bool flat  ) /*const*/
{
	QColor highlight(colorBottom);
	QColor subh1(colorTop);
	QColor subh2(colorTop);

	QColor topgrad(colorTop);
	QColor botgrad(colorBottom);


	if ( flat && !sunken )
		p->fillRect(x, y, w, h, colorTop);
	else {
		int i  = 0;
		int x1 = x;
		int y1 = y;
		int x2 = x + w - 1;
		int y2 = y + h - 1;
		if ( horz )
			x2 = x2;
		else
			y2 = y2;

#define DRAWLINE if (horz) \
                    p->drawLine( x1, y1+i, x2, y1+i ); \
                 else \
                    p->drawLine( x1+i, y1, x1+i, y2 ); \
                 i++;

		// Gradient:
		int ng = (horz ? h : w); // how many lines for the gradient?

		int h1, h2, s1, s2, v1, v2;
		if ( !sunken ) {
			topgrad.hsv( &h1, &s1, &v1 );
			botgrad.hsv( &h2, &s2, &v2 );
		} else {
			botgrad.hsv( &h1, &s1, &v1 );
			topgrad.hsv( &h2, &s2, &v2 );
		}

		if ( ng > 1 ) {
			for ( int j =0; j < ng; j++ ) {
				p->setPen( QColor( h1 + ((h2-h1)*j)/(ng-1),
						   s1 + ((s2-s1)*j)/(ng-1),
						   v1 + ((v2-v1)*j)/(ng-1),  QColor::Hsv ) );
				DRAWLINE;
			}
		} else if ( ng == 1 ) {
			p->setPen( QColor((h1+h2)/2, (s1+s2)/2, (v1+v2)/2, QColor::Hsv) );
			DRAWLINE;
		}
	}
}

void Item::drawExpander(QPainter *painter, int x, int y, const QColor &background, bool expand, Basket *basket)
{
	// If the current style is a KStyle, use it to draw the expander (plus or minus):
	if (dynamic_cast<KStyle*>(&(kapp->style())) != NULL) {
		// Set the 4 rounded corners background to background color:
		QColorGroup cg(basket->colorGroup());
		cg.setColor(QColorGroup::Base, background);

		// Fill the inside of the expander in white, typically:
		QBrush brush(KGlobalSettings::baseColor());
		painter->fillRect(x, y, 9, 9, brush);

		// Draw it:
		((KStyle&)(kapp->style())).drawKStylePrimitive( KStyle::KPE_ListViewExpander,
		painter,
		basket->viewport(),
		QRect(x, y, 9, 9),
		cg,
		(expand ? QStyle::Style_On : QStyle::Style_Off) );
	// Else, QStyle does not provide easy way to do so (if it's doable at all...)
	// So, I'm drawing it myself my immitating Plastik (pretty style)...
	// After all, the item/group handles are all non-QStyle aware so that doesn't matter if the expander is a custom one too.
	} else {
		int width  = EXPANDER_WIDTH;
		int height = EXPANDER_HEIGHT;
		const QColorGroup &cg = basket->colorGroup();

		// Fill white area:
		painter->fillRect(x + 1, y + 1, width - 2, height - 2, cg.base());
		// Draw contour lines:
		painter->setPen(cg.dark());
		painter->drawLine(x + 2,         y,              x + width - 3, y);
		painter->drawLine(x + 2,         y + height - 1, x + width - 3, y + height - 1);
		painter->drawLine(x,             y + 2,          x,             y + height - 3);
		painter->drawLine(x + width - 1, y + 2,          x + width - 1, y + height - 3);
		// Draw edge points:
		painter->drawPoint(x + 1,         y + 1);
		painter->drawPoint(x + width - 2, y + 1);
		painter->drawPoint(x + 1,         y + height - 2);
		painter->drawPoint(x + width - 2, y + height - 2);
		// Draw anti-aliased points:
		painter->setPen(Utilities::mixColor(cg.dark(), background));
		painter->drawPoint(x + 1,         y);
		painter->drawPoint(x + width - 2, y);
		painter->drawPoint(x,             y + 1);
		painter->drawPoint(x + width - 1, y + 1);
		painter->drawPoint(x,             y + height - 2);
		painter->drawPoint(x + width - 1, y + height - 2);
		painter->drawPoint(x + 1,         y + height - 1);
		painter->drawPoint(x + width - 2, y + height - 1);
		// Draw plus / minus:
		painter->setPen(cg.text());
		painter->drawLine(x + 2, y + height / 2, x + width - 3, y + height / 2);
		if (expand)
			painter->drawLine(x + width / 2, y + 2, x + width / 2, y + height - 3);
	}
}

QColor expanderBackground(int height, int y, const QColor &foreground)
{
/*	// We will divide height per two, substract one and use that below a division bar:
	if (height <= 3)
		return foreground;

	QColor dark     = foreground.dark(110);  // 1/1.1 of brightness
	QColor light    = foreground.light(150); // 50% brighter
//	QColor midLight = foreground.light(105); // 5% brighter

	int h1, h2, s1, s2, v1, v2;
	light.hsv( &h1, &s1, &v1 );
	dark.hsv(  &h2, &s2, &v2 );
	int ng = height / 2;
	return QColor( h1 + ((h2-h1)*y)/(ng-1),
	               s1 + ((s2-s1)*y)/(ng-1),
	               v1 + ((v2-v1)*y)/(ng-1), QColor::Hsv );*/

	// We will divide height per two, substract one and use that below a division bar:
	// To avoid division by zero error, height should be bigger than 3.
	// And to avoid y errors or if y is on the borders, we return the border color: the background color.
	if (height <= 3 || y <= 0 || y >= height - 1)
		return foreground;

	QColor dark     = foreground.dark(110);  // 1/1.1 of brightness
	QColor light    = foreground.light(150); // 50% brighter
//	QColor midLight = foreground.light(105); // 5% brighter

	int h1, h2, s1, s2, v1, v2;
	int ng;
	if (y <= (height-2)/2) {
		light.hsv( &h1, &s1, &v1 );
		dark.hsv(  &h2, &s2, &v2 );
		ng = (height-2)/2;
		y -= 1;
	} else {
		dark.hsv( &h1, &s1, &v1 );
		foreground.hsv(  &h2, &s2, &v2 );
		ng = (height-2)-(height-2)/2;
		y -= 1 + (height-2)/2;
	}
	return QColor( h1 + ((h2-h1)*y)/(ng-1),
	               s1 + ((s2-s1)*y)/(ng-1),
	               v1 + ((v2-v1)*y)/(ng-1), QColor::Hsv );
}

void Item::drawHandle(QPainter *painter, int x, int y, int width, int height, const QColor &background, const QColor &foreground)
{
	QPen backgroundPen(background);
	QPen foregroundPen(foreground);

	QColor dark     = foreground.dark(110);  // 1/1.1 of brightness
	QColor light    = foreground.light(150); // 50% brighter
//	QColor midLight = foreground.light(105); // 5% brighter

	// Draw the surrounding rectangle:
	painter->setPen(foregroundPen);
	painter->drawLine(0,         0,          width - 1, 0);
	painter->drawLine(0,         0,          0,         height - 1);
	painter->drawLine(width - 1, 0,          width - 1, height - 1);
	painter->drawLine(0,         height - 1, width - 1, height - 1);

	// Draw the gradients:
	drawGradient( painter, light, dark/*midLight*/, 1 + x, 1 + y,                width-2, (height-2)/2,            /*sunken=*/false, /*horz=*/true, /*flat=*/false  );
	drawGradient( painter, dark,  foreground,       1 + x, 1 + y + (height-2)/2, width-2, (height-2)-(height-2)/2, /*sunken=*/false, /*horz=*/true, /*flat=*/false  );

	// Round the top corner with background color:
	painter->setPen(backgroundPen);
	painter->drawLine(0, 0, 0, 3);
	painter->drawLine(1, 0, 3, 0);
	painter->drawPoint(1, 1);
	// Round the bottom corner with background color:
	painter->drawLine(0, height-1, 0, height-4);
	painter->drawLine(1, height-1, 3, height-1);
	painter->drawPoint(1, height-2);

	// Surrounding line of the rounded top-left corner:
	painter->setPen(foregroundPen);
	painter->drawLine(1, 2, 1, 3);
	painter->drawLine(2, 1, 3, 1);

	// Anti-aliased rounded top corner (1/2):
	painter->setPen(Utilities::mixColor(foreground, background));
	painter->drawPoint(0, 3);
	painter->drawPoint(3, 0);
	// Anti-aliased rounded bottom corner:
	painter->drawPoint(0, height - 4);
	painter->drawPoint(3, height - 1);
	// Anti-aliased rounded top corner (2/2):
	painter->setPen(Utilities::mixColor(foreground, light));
	painter->drawPoint(2, 2);

	// Draw the grips:
	int xGrips             = 4;
	int marginedHeight = (height * 80 / 100); // 10% empty on top, and 10% empty on bottom, so 20% of the height should be empty of any grip, and 80% should be in the grips
	int nbGrips            = (marginedHeight - 3) / 6;
	if (nbGrips < 2)
		nbGrips = 2;
	int yGrips             = (height + 1 - nbGrips * 6 - 3) / 2; // +1 to avoid rounding errors, -nbGrips*6-3 the size of the grips
	QColor darker  = foreground.dark(130);
	QColor lighter = foreground.light(130);
	for (int i = 0; i < nbGrips; ++i) {
		/// Dark color:
		painter->setPen(darker);
		// Top-left point:
		painter->drawPoint(xGrips,     yGrips);
		painter->drawPoint(xGrips + 1, yGrips);
		painter->drawPoint(xGrips,     yGrips + 1);
		// Bottom-right point:
		painter->drawPoint(xGrips + 4, yGrips + 3);
		painter->drawPoint(xGrips + 5, yGrips + 3);
		painter->drawPoint(xGrips + 4, yGrips + 4);
		/// Light color:
		painter->setPen(lighter);
		// Top-left point:
		painter->drawPoint(xGrips + 1, yGrips + 1);
		// Bottom-right point:
		painter->drawPoint(xGrips + 5, yGrips + 4);
		yGrips += 6;
	}
	// The remaining point:
	painter->setPen(darker);
	painter->drawPoint(xGrips,     yGrips);
	painter->drawPoint(xGrips + 1, yGrips);
	painter->drawPoint(xGrips,     yGrips + 1);
	painter->setPen(lighter);
	painter->drawPoint(xGrips + 1, yGrips + 1);
}

void Item::drawResizer(QPainter *painter, int x, int y, int width, int height, const QColor &background, const QColor &foreground, bool rounded)
{
	QPen backgroundPen(background);
	QPen foregroundPen(foreground);

	QColor dark     = foreground.dark(110);  // 1/1.1 of brightness
	QColor light    = foreground.light(150); // 50% brighter
	QColor midLight = foreground.light(105); // 5% brighter

	// Draw the surrounding rectangle:
	painter->setPen(foregroundPen);
	painter->drawRect(0, 0, width, height);
// 	painter->drawLine(0,         0,          width - 1, 0);
// 	painter->drawLine(0,         0,          0,         height - 1);
// 	painter->drawLine(width - 1, 0,          width - 1, height - 1);
// 	painter->drawLine(0,         height - 1, width - 1, height - 1);

	// Draw the gradients:
	drawGradient( painter, light, dark/*midLight*/, 1 + x, 1 + y,                width-2, (height-2)/2,            /*sunken=*/false, /*horz=*/true, /*flat=*/false  );
	drawGradient( painter, dark,  foreground,       1 + x, 1 + y + (height-2)/2, width-2, (height-2)-(height-2)/2, /*sunken=*/false, /*horz=*/true, /*flat=*/false  );

	if (rounded) {
		// Round the top corner with background color:
		painter->setPen(backgroundPen);
		painter->drawLine(width - 1, 0, width - 3, 0);
		painter->drawLine(width - 1, 1, width - 1, 2);
		painter->drawPoint(width - 2, 1);
		// Round the bottom corner with background color:
		painter->drawLine(width - 1, height - 1, width - 1, height - 4);
		painter->drawLine(width - 2, height - 1, width - 4, height - 1);
		painter->drawPoint(width - 2, height-2);

		// Surrounding line of the rounded top-left corner:
		painter->setPen(foregroundPen);
		painter->drawLine(width-2, 2, width-2, 3);
		painter->drawLine(width-3, 1, width-4, 1);

		// Anti-aliased rounded top corner (1/2):
		painter->setPen(Utilities::mixColor(foreground, background));
		painter->drawPoint(width - 1, 3);
		painter->drawPoint(width - 4, 0);
		// Anti-aliased rounded bottom corner:
		painter->drawPoint(width - 1, height - 4);
		painter->drawPoint(width - 4, height - 1);
		// Anti-aliased rounded top corner (2/2):
		painter->setPen(Utilities::mixColor(foreground, light));
		painter->drawPoint(width - 3, 2);
	}

	// Draw the arows:
	int xArrow  = 2;
	int hMargin = 9;
	int countArrows = (height >= hMargin*4 + 6*3 ? 3 : (height >= hMargin*3 + 6*2 ? 2 : 1));
	QColor darker  = foreground.dark(130);
	QColor lighter = foreground.light(130);
	for (int i = 0; i < countArrows; ++i) {
		int yArrow;
		switch (countArrows) {
			case 1: yArrow = (height-6) / 2;                                                        break;
			case 2: yArrow = (i == 1 ? hMargin : height - hMargin - 6);                             break;
			case 3: yArrow = (i == 1 ? hMargin : (i == 2 ? (height-6) / 2 : height - hMargin - 6)); break;
		}
		/// Dark color:
		painter->setPen(darker);
		// Left arrow:
		painter->drawLine(xArrow, yArrow + 2, xArrow + 2, yArrow);
		painter->drawLine(xArrow, yArrow + 2, xArrow + 2, yArrow + 4);
		// Right arrow:
		painter->drawLine(width - 1 - xArrow, yArrow + 2, width - 1 - xArrow - 2, yArrow);
		painter->drawLine(width - 1 - xArrow, yArrow + 2, width - 1 - xArrow - 2, yArrow + 4);
		/// Light color:
		painter->setPen(lighter);
		// Left arrow:
		painter->drawLine(xArrow, yArrow + 2 + 1, xArrow + 2, yArrow + 1);
		painter->drawLine(xArrow, yArrow + 2 + 1, xArrow + 2, yArrow + 4 + 1);
		// Right arrow:
		painter->drawLine(width - 1 - xArrow, yArrow + 2 + 1, width - 1 - xArrow - 2, yArrow + 1);
		painter->drawLine(width - 1 - xArrow, yArrow + 2 + 1, width - 1 - xArrow - 2, yArrow + 4 + 1);
	}
}

void Item::drawInactiveResizer(QPainter *painter, int x, int y, int height, const QColor &background, bool column)
{
	// If background color is too dark, we compute a lighter color instead of a darker:
	QColor darkBgColor = (tooDark(background) ? background.light(120) : background.dark(105));
	if (column) {
		int halfWidth = RESIZER_WIDTH / 2;
		drawGradient(painter, darkBgColor, background,  x,         y, halfWidth,                 height, /*sunken=*/false, /*horz=*/false, /*flat=*/false);
		drawGradient(painter, background,  darkBgColor, halfWidth, y, RESIZER_WIDTH - halfWidth, height, /*sunken=*/false, /*horz=*/false, /*flat=*/false);
	} else
		drawGradient(painter, darkBgColor, background,  x,         y, RESIZER_WIDTH,             height, /*sunken=*/false, /*horz=*/false, /*flat=*/false );
}


#include <qimage.h>
#include <kimageeffect.h>

/* type: 1: topLeft
 *       2: bottomLeft
 *       3: topRight
 *       4: bottomRight
 *       5: fourCorners
 *       6: itemInsideAndOutsideCorners
 * (x,y) relate to the painter origin
 * (width,height) only used for 5:fourCorners type
 */
void Item::drawRoundings(QPainter *painter, int x, int y, int type, int width, int height)
{
	int right;

	switch (type) {
		case 1:
			x += this->x();
			y += this->y();
			basket()->blendBackground(*painter, QRect(x, y,     /*3*/4, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x, y + 1, 2, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x, y + 2, 1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x, y + 3, 1, 1), this->x(), this->y());
			break;
		case 2:
			x += this->x();
			y += this->y();
			basket()->blendBackground(*painter, QRect(x, y - 1,     1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x, y,     1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x, y + 1, 2, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x, y + 2, /*3*/4, 1), this->x(), this->y());
			break;
		case 3:
			right = rightLimit();
			x += right;
			y += this->y();
			basket()->blendBackground(*painter, QRect(x - 1, y,     /*3*/4, 1), right, this->y());
			basket()->blendBackground(*painter, QRect(x + 1, y + 1, 2, 1), right, this->y());
			basket()->blendBackground(*painter, QRect(x + 2, y + 2, 1, 1), right, this->y());
			basket()->blendBackground(*painter, QRect(x + 2, y + 3, 1, 1), right, this->y());
			break;
		case 4:
			right = rightLimit();
			x += right;
			y += this->y();
			basket()->blendBackground(*painter, QRect(x + 2, y - 1, 1, 1), right, this->y());
			basket()->blendBackground(*painter, QRect(x + 2, y,     1, 1), right, this->y());
			basket()->blendBackground(*painter, QRect(x + 1, y + 1, 2, 1), right, this->y());
			basket()->blendBackground(*painter, QRect(x - 1,     y + 2, /*3*/4, 1), right, this->y());
			break;
		case 5:
			// First make sure the corners are white (depending on the widget style):
			painter->setPen(basket()->backgroundColor());
			painter->drawPoint(x,             y);
			painter->drawPoint(x + width - 1, y);
			painter->drawPoint(x + width - 1, y + height - 1);
			painter->drawPoint(x,             y + height - 1);
			// And then blend corners:
			x += this->x();
			y += this->y();
			basket()->blendBackground(*painter, QRect(x,             y,              1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x + width - 1, y,              1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x + width - 1, y + height - 1, 1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x,             y + height - 1, 1, 1), this->x(), this->y());
			break;
		case 6:
			x += this->x();
			y += this->y();
			//if (!isSelected()) {
				// Inside left corners:
			basket()->blendBackground(*painter, QRect(x + HANDLE_WIDTH + 1, y + 1,          1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x + HANDLE_WIDTH,     y + 2,          1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x + HANDLE_WIDTH + 1, y + height - 2, 1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x + HANDLE_WIDTH,     y + height - 3, 1, 1), this->x(), this->y());
				// Inside right corners:
			basket()->blendBackground(*painter, QRect(x + width - 4,        y + 1,          1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x + width - 3,        y + 2,          1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x + width - 4,        y + height - 2, 1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x + width - 3,        y + height - 3, 1, 1), this->x(), this->y());
			//}
			// Outside right corners:
			basket()->blendBackground(*painter, QRect(x + width - 1,        y,              1, 1), this->x(), this->y());
			basket()->blendBackground(*painter, QRect(x + width - 1,        y + height - 1, 1, 1), this->x(), this->y());
			break;
	}
}

/// Blank Spaces Drawing:

void Item::setOnTop(bool onTop)
{
	m_onTop = onTop;

	Item *item = firstChild();
	while (item) {
		item->setOnTop(onTop);
		item = item->next();
	}
}

void Item::recomputeAreas()
{
	m_areas.clear();
	m_areas.append(rect());
	if (hasResizer())
		m_areas.append(resizerRect());

	Item *item = basket()->firstItem();
	bool itemIsAfterThis = false;
	while (item) {
		itemIsAfterThis = recomputeAreas(item, itemIsAfterThis);
		item = item->next();
	}
}

void substractRectOnAreas(const QRect &rectToSubstract, QValueList<QRect> &areas, bool andRemove = true)
{
	for (QValueList<QRect>::iterator it = areas.begin(); it != areas.end(); ) {
		QRect &rect = *it;
		// Split the rectangle if it intersects with rectToSubstract:
		if (rect.intersects(rectToSubstract)) {
			// Create the top rectangle:
			if (rectToSubstract.top() > rect.top()) {
				areas.insert(it, QRect(rect.left(), rect.top(), rect.width(), rectToSubstract.top() - rect.top()));
				rect.setTop(rectToSubstract.top());
			}
			// Create the bottom rectangle:
			if (rectToSubstract.bottom() < rect.bottom()) {
				areas.insert(it, QRect(rect.left(), rectToSubstract.bottom() + 1, rect.width(), rect.bottom() - rectToSubstract.bottom()));
				rect.setBottom(rectToSubstract.bottom());
			}
			// Create the left rectangle:
			if (rectToSubstract.left() > rect.left()) {
				areas.insert(it, QRect(rect.left(), rect.top(), rectToSubstract.left() - rect.left(), rect.height()));
				rect.setLeft(rectToSubstract.left());
			}
			// Create the right rectangle:
			if (rectToSubstract.right() < rect.right()) {
				areas.insert(it, QRect(rectToSubstract.right() + 1, rect.top(), rect.right() - rectToSubstract.right(), rect.height()));
				rect.setRight(rectToSubstract.right());
			}
			// Remove the rectangle if it's entirely contained:
			if (andRemove && rectToSubstract.contains(rect))
				it = areas.remove(it);
			else
				++it;
		} else
			++it;
	}
}

bool Item::recomputeAreas(Item *item, bool itemIsAfterThis)
{
	if (item == this)
		itemIsAfterThis = true;
	// Only compute overlapping of items AFTER this, or ON TOP this:
	else if ( item->matching() && itemIsAfterThis && (!isOnTop() || (isOnTop() && item->isOnTop())) || (!isOnTop() && item->isOnTop()) ) {
		substractRectOnAreas(item->rect(), m_areas);
		if (item->hasResizer())
			substractRectOnAreas(item->resizerRect(), m_areas);
	}

	if (item->isGroup()) {
		Item *child = item->firstChild();
		bool first = true;
		while (child) {
			if ((item->showSubItems() || first) && item->matching())
				itemIsAfterThis = recomputeAreas(child, itemIsAfterThis);
			child = child->next();
			first = false;
		}
	}

	return itemIsAfterThis;
}

void Item::getGradientColors(const QColor &originalBackground, QColor *colorTop, QColor *colorBottom)
{
	bool wasTooDark = tooDark(originalBackground);
	if (wasTooDark) {
		*colorTop    = originalBackground;
		*colorBottom = originalBackground.light(120);
	} else {
		*colorTop    = originalBackground.dark(105);
		*colorBottom = originalBackground;
	}
}


/* Drawing policy:
 * ==============
 * - Draw the item on a pixmap and then draw the pixmap on screen is faster and flicker-free, rather than drawing directly on screen
 * - The next time the pixmap can be directly redrawn on screen without (relatively low, for small texts) time-consuming text-drawing
 * - To keep memory footprint low, we can destruct the bufferPixmap because redrawing it offscreen and applying it onscreen is nearly as fast as just drawing the pixmap onscreen
 * - But as drawing the pixmap offscreen is little time consuming we can keep last visible items buffered and then the redraw of the entire window is INSTANTANEOUS
 * - We keep bufferized item/group draws BUT NOT the resizer: such objects are small and fast to draw, so we don't complexify code for that
 */
void Item::draw(QPainter *painter, const QRect &clipRect)
{
	if (!matching())
		return;

	/** Paint childs: */
	if (isGroup()) {
		Item *child = firstChild();
		bool first = true;
		while (child) {
			if ((showSubItems() || first) && child->matching())
				child->draw(painter, clipRect);
			child = child->next();
			first = false;
		}
	}

	QRect myRect(x(), y(), width(), height());
	/** Paint the resizer if needed: */
	if (hasResizer()) {
		int right = rightLimit();
		QRect resizerRect(right, y(), RESIZER_WIDTH, resizerHeight());
		if (resizerRect.intersects(clipRect)) {
			// Prepare to draw the resizer:
			QPixmap pixmap(RESIZER_WIDTH, resizerHeight());
			QPainter painter2(&pixmap);
			// Draw gradient or resizer:
			if (m_hovered && m_hoveredZone == Resizer) {
				QColor baseColor(basket()->backgroundColor());
				QColor highColor(KGlobalSettings::highlightColor());
				drawResizer(&painter2, 0, 0, RESIZER_WIDTH, resizerHeight(), baseColor, highColor, /*rounded=*/!isColumn());
				if (!isColumn()) {
					drawRoundings(&painter2, RESIZER_WIDTH - 3, 0,                   /*type=*/3);
					drawRoundings(&painter2, RESIZER_WIDTH - 3, resizerHeight() - 3, /*type=*/4);
				}
			} else {
				drawInactiveResizer(&painter2, /*x=*/0, /*y=*/0, /*height=*/resizerHeight(), basket()->backgroundColor(), isColumn());
 				basket()->blendBackground(painter2, resizerRect);
			}
			// Draw inserter:
			if (basket()->inserterShown() && resizerRect.intersects(basket()->inserterRect()))
				basket()->drawInserter(painter2, right, y());
			// Draw selection rect:
			if (basket()->isSelecting() && resizerRect.intersects(basket()->selectionRect())) {
				QRect selectionRect = basket()->selectionRect();
				selectionRect.moveBy(-right, -y());
				QRect selectionRectInside(selectionRect.x() + 1, selectionRect.y() + 1, selectionRect.width() - 2, selectionRect.height() - 2);
				if (selectionRectInside.width() > 0 && selectionRectInside.height() > 0) {
					QColor insideColor = basket()->selectionRectInsideColor();
					QColor darkInsideColor(insideColor.dark(105));
					//painter2.fillRect(selectionRectInside, insideColor);
					painter2.setClipRect(selectionRectInside);
					if (isColumn()) {
						int halfWidth = RESIZER_WIDTH / 2;
						drawGradient(&painter2, darkInsideColor, insideColor,     0,         0, halfWidth,               resizerHeight(), /*sunken=*/false, /*horz=*/false, /*flat=*/false );
						drawGradient(&painter2, insideColor,     darkInsideColor, halfWidth, 0, RESIZER_WIDTH-halfWidth, resizerHeight(), /*sunken=*/false, /*horz=*/false, /*flat=*/false );
					} else
						drawGradient( &painter2, darkInsideColor, insideColor, 0, 0, RESIZER_WIDTH, resizerHeight(), /*sunken=*/false, /*horz=*/false, /*flat=*/false  );
						painter2.setClipping(false);
						selectionRectInside.moveBy(right, y());
						basket()->blendBackground(painter2, selectionRectInside, right, y(), false);
				}
				painter2.setPen(KGlobalSettings::highlightColor().dark());
				painter2.drawRect(selectionRect);
				painter2.setPen(Utilities::mixColor(KGlobalSettings::highlightColor().dark(), basket()->backgroundColor()));
				painter2.drawPoint(selectionRect.topLeft());
				painter2.drawPoint(selectionRect.topRight());
				painter2.drawPoint(selectionRect.bottomLeft());
				painter2.drawPoint(selectionRect.bottomRight());
			}
			// Draw on screen:
			painter2.end();
			/** Compute visible areas: */
			if ( ! m_computedAreas )
				recomputeAreas();
			if (m_areas.isEmpty())
				return;
			for (QValueList<QRect>::iterator it = m_areas.begin(); it != m_areas.end(); ++it) {
				QRect &rect = *it;
				painter->drawPixmap(rect.x(), rect.y(), pixmap, rect.x() - right, rect.y() - y(), rect.width(), rect.height());
			}
		}
	}

	/** Then, draw the item/group ONLY if needed: */
	if ( ! myRect.intersects(clipRect) )
		return;

	/** Compute visible areas: */
	if ( ! m_computedAreas )
		recomputeAreas();
	if (m_areas.isEmpty())
		return;

	/** Directly draw pixmap on screen if it is already buffered: */
	if (isBufferized()) {
		drawBufferOnScreen(painter, m_bufferedPixmap);
		return;
	}

	/** Initialise buffer painter: */
	m_bufferedPixmap.resize(width(), height());
	QPainter painter2(&m_bufferedPixmap);

	/** Initialise colors: */
	QColor baseColor(basket()->backgroundColor());
	QColor highColor(KGlobalSettings::highlightColor());
	QColor midColor = Utilities::mixColor(baseColor, highColor);

	/** Initialise brushs and pens: */
	QBrush baseBrush(baseColor);
	QBrush highBrush(highColor);
	QPen   basePen(baseColor);
	QPen   highPen(highColor);
	QPen   midPen(midColor);

	/** Figure out the state of the item: */
	bool hovered = m_hovered && m_hoveredZone != TopInsert && m_hoveredZone != BottomInsert && m_hoveredZone != Resizer;

	/** And then draw the group: */
	if (isGroup()) {
		// Draw background or handle:
		if (hovered) {
			drawHandle(&painter2, 0, 0, width(), height(), baseColor, highColor);
			drawRoundings(&painter2, 0, 0,            /*type=*/1);
			drawRoundings(&painter2, 0, height() - 3, /*type=*/2);
		} else {
			painter2.fillRect(0, 0, width(), height(), baseBrush);
			basket()->blendBackground(painter2, myRect, -1, -1, /*opaque=*/true);
		}

		// Draw expander:
		int yExp = yExpander();
		drawExpander(&painter2, ITEM_MARGIN, yExp, (hovered ? expanderBackground(height(), yExp+EXPANDER_HEIGHT/2, highColor) : baseColor), m_isFolded, basket());
		// Draw expander rounded edges:
		if (hovered) {
			QColor color1 = expanderBackground(height(), yExp,                       highColor);
			QColor color2 = expanderBackground(height(), yExp + EXPANDER_HEIGHT - 1, highColor);
			painter2.setPen(color1);
			painter2.drawPoint(ITEM_MARGIN,         yExp);
			painter2.drawPoint(ITEM_MARGIN + 9 - 1, yExp);
			painter2.setPen(color2);
			painter2.drawPoint(ITEM_MARGIN,         yExp + 9 - 1);
			painter2.drawPoint(ITEM_MARGIN + 9 - 1, yExp + 9 - 1);
		} else
			drawRoundings(&painter2, ITEM_MARGIN, yExp, /*type=*/5, 9, 9);

		// Draw on screen:
		painter2.end();
		drawBufferOnScreen(painter, m_bufferedPixmap);
//		unbufferize();
		return;
	}

	/** Or draw the item: */
	// What are the background colors:
	QColor background = basket()->backgroundColor();
	if (isSelected())
		if (m_computedState.backgroundColor().isValid())
			background = Utilities::mixColor(Utilities::mixColor(m_computedState.backgroundColor(), KGlobalSettings::highlightColor()), KGlobalSettings::highlightColor());
		else
			background = KGlobalSettings::highlightColor();
	else if (m_computedState.backgroundColor().isValid())
		background = m_computedState.backgroundColor();
	QColor bgColor;
	QColor darkBgColor;
	getGradientColors(background, &darkBgColor, &bgColor);
	// Draw background (color, gradient and pixmap):
	drawGradient( &painter2, bgColor, darkBgColor, 0, 0, width(), height(), /*sunken=*/!hovered, /*horz=*/true, /*flat=*/false );
	if (!hovered) {
		painter2.setPen(Utilities::mixColor(bgColor, darkBgColor));
		painter2.drawLine(0, height() - 1, width(), height() - 1);
	}
	basket()->blendBackground(painter2, myRect);

	if (hovered) {
		// Top/Bottom lines:
		painter2.setPen(highPen);
		painter2.drawLine(0, height() - 1, width(), height() - 1);
		painter2.drawLine(0, 0,            width(), 0);
		// The handle:
		drawHandle(&painter2, 0, 0, HANDLE_WIDTH, height(), baseColor, highColor);
		drawRoundings(&painter2, 0, 0,            /*type=*/1);
		drawRoundings(&painter2, 0, height() - 3, /*type=*/2);
		// Round handle-right-side border:
		painter2.setPen(highPen);
		painter2.drawPoint(HANDLE_WIDTH, 1);
		painter2.drawPoint(HANDLE_WIDTH, height() - 2);
		// Light handle top-right round corner:
		painter2.setPen(QPen(highColor.light(150)));
		painter2.drawPoint(HANDLE_WIDTH - 1, 1);
		// Handle anti-aliased rounded handle-right-side corners:
		QColor insideMidColor = Utilities::mixColor(bgColor, highColor);
		painter2.setPen(insideMidColor);
		// Left inside round corners:
		painter2.drawPoint(HANDLE_WIDTH + 1, 1);
		painter2.drawPoint(HANDLE_WIDTH,     2);
		painter2.drawPoint(HANDLE_WIDTH + 1, height() - 2);
		painter2.drawPoint(HANDLE_WIDTH,     height() - 3);
		// Right inside round corners:
		painter2.drawPoint(width() - 4, 1);
		painter2.drawPoint(width() - 3, 2);
		painter2.drawPoint(width() - 4, height() - 2);
		painter2.drawPoint(width() - 3, height() - 3);
		// Right rounded edge:
		painter2.setPen(highPen);
		painter2.fillRect(width() - 2, 0, 2, height(), highBrush);
		painter2.drawPoint(width() - 3, 1);
		painter2.drawPoint(width() - 3, height() - 2);
		// Right anti-aliased rounded edge:
		painter2.setPen(midPen);
		painter2.drawPoint(width() - 1, 0);
		painter2.drawPoint(width() - 1, height() - 1);
		// Blend background pixmap:
		drawRoundings(&painter2, 0, 0, /*type=*/6, width(), height());
	}

	// Draw the Emblems:
	int yIcon = (height() - EMBLEM_SIZE) / 2;
	int xIcon = HANDLE_WIDTH + ITEM_MARGIN;
	for (State::List::Iterator it = m_states.begin(); it != m_states.end(); ++it) {
		if (!(*it)->emblem().isEmpty()) {
			QPixmap stateEmblem = kapp->iconLoader()->loadIcon((*it)->emblem(), KIcon::NoGroup, 16, KIcon::DefaultState, 0L, false);
			painter2.drawPixmap(xIcon, yIcon, stateEmblem);
			xIcon += ITEM_MARGIN + EMBLEM_SIZE;
		}
	}

	// Determine the colors (for the richText drawing) and the text color (for the tags arrow too):
	QColorGroup cg(basket()->colorGroup());
	cg.setColor(QColorGroup::Text,       (m_computedState.textColor().isValid() ? m_computedState.textColor() : basket()->textColor()) );
	cg.setColor(QColorGroup::Background, bgColor);
	if (isSelected())
		cg.setColor(QColorGroup::Text, KGlobalSettings::highlightedTextColor());

	// Draw the Tags Arrow:
	if (hovered) {
		QColor textColor = cg.color(QColorGroup::Text);
		QColor light     = Utilities::mixColor(textColor, bgColor);
		QColor mid       = Utilities::mixColor(textColor, light);
		painter2.setPen(light);//QPen(basket()->colorGroup().dark().light(150)));
		painter2.drawLine(xIcon,      yIcon + 6, xIcon + 4, yIcon + 6);
		painter2.setPen(mid);//QPen(basket()->colorGroup().dark()));
		painter2.drawLine(xIcon + 1,  yIcon + 7, xIcon + 3, yIcon + 7);
		painter2.setPen(textColor);//QPen(basket()->colorGroup().foreground()));
		painter2.drawPoint(xIcon + 2, yIcon + 8);
	} else if (m_haveInvisibleTags) {
		painter2.setPen(cg.color(QColorGroup::Text)/*QPen(basket()->colorGroup().foreground())*/);
		painter2.drawPoint(xIcon,     yIcon + 7);
		painter2.drawPoint(xIcon + 2, yIcon + 7);
		painter2.drawPoint(xIcon + 4, yIcon + 7);
	}

	// Draw content:
	painter2.translate(contentX(), ITEM_MARGIN);
	painter2.setFont( m_computedState.font(painter2.font()) );
	m_content->paint(&painter2, width() - contentX() - ITEM_MARGIN, height() - 2*ITEM_MARGIN, cg, !m_computedState.textColor().isValid(), isSelected(), hovered);

	// Draw on screen:
	painter2.end();
	drawBufferOnScreen(painter, m_bufferedPixmap);

//	unbufferize();
}

void Item::drawBufferOnScreen(QPainter *painter, const QPixmap &contentPixmap)
{
	for (QValueList<QRect>::iterator it = m_areas.begin(); it != m_areas.end(); ++it) {
		QRect &rect = *it;
		if (rect.x() >= x() + width()) // It's a rect of the resizer, don't draw it!
			continue;
		// If the inserter is above the item, draw it, BUT NOT in the buffer pixmap,
		// we copy the rectangle in a new pixmap, apply the inserter and then draw this new pixmap on screen:
		if (  (basket()->inserterShown() && rect.intersects(basket()->inserterRect()))  ||
		      (basket()->isSelecting()   && rect.intersects(basket()->selectionRect()))  ) {
			QPixmap pixmap3(rect.width(), rect.height());
			QPainter painter3(&pixmap3);
			painter3.drawPixmap(0, 0, contentPixmap, rect.x() - x(), rect.y() - y(), rect.width(), rect.height());
			// Draw inserter:
			if (basket()->inserterShown() && rect.intersects(basket()->inserterRect()))
				basket()->drawInserter(painter3, rect.x(), rect.y());
			// Draw selection rect:
			if (basket()->isSelecting() && rect.intersects(basket()->selectionRect())) {
				QRect selectionRect = basket()->selectionRect();
				selectionRect.moveBy(-rect.x(), -rect.y());

				QRect selectionRectInside(selectionRect.x() + 1, selectionRect.y() + 1, selectionRect.width() - 2, selectionRect.height() - 2);
				if (selectionRectInside.width() > 0 && selectionRectInside.height() > 0) {
					bufferizeSelectionPixmap();
					selectionRectInside.moveBy(rect.x(), rect.y());
					QRect rectToPaint = rect.intersect(selectionRectInside);
					rectToPaint.moveBy(-x(), -y());
					painter3.drawPixmap(rectToPaint.topLeft() + QPoint(x(), y()) - rect.topLeft(), m_bufferedSelectionPixmap, rectToPaint);
					//blendBackground(painter2, selectionRectInside, rect.x(), rect.y(), true, &m_selectedBackgroundPixmap);
				}

				painter3.setPen(KGlobalSettings::highlightColor().dark());
				painter3.drawRect(selectionRect);
				if (isGroup())
					painter3.setPen(Utilities::mixColor(KGlobalSettings::highlightColor().dark(), basket()->backgroundColor()));
				else {
					// What are the background colors:
					QColor bgColor = basket()->backgroundColor();
					if (isSelected())
						bgColor = (m_computedState.backgroundColor().isValid() ? Utilities::mixColor(Utilities::mixColor(m_computedState.backgroundColor(), KGlobalSettings::highlightColor()), KGlobalSettings::highlightColor()) : KGlobalSettings::highlightColor());
					else if (m_computedState.backgroundColor().isValid())
						bgColor = m_computedState.backgroundColor();
					painter3.setPen(Utilities::mixColor(KGlobalSettings::highlightColor().dark(), bgColor));
				}
				painter3.drawPoint(selectionRect.topLeft());
				painter3.drawPoint(selectionRect.topRight());
				painter3.drawPoint(selectionRect.bottomLeft());
				painter3.drawPoint(selectionRect.bottomRight());
			}
			painter3.end();
			painter->drawPixmap(rect.x(), rect.y(), pixmap3);
		// Else, draw the rect pixmap directly on screen:
		} else
			painter->drawPixmap(rect.x(), rect.y(), contentPixmap, rect.x() - x(), rect.y() - y(), rect.width(), rect.height());
	}
}

void Item::setContent(ItemContent *content)
{
	m_content = content;
	// TODO
}

/*const */State::List& Item::states() const
{
	return (State::List&)m_states;
}

void Item::addState(State *state, bool orReplace)
{
	if (!content())
		return;

	Tag *tag = state->parentTag();
	State::List::iterator itStates = m_states.begin();
	// Browse all tags, see if the item has it, increment itSates if yes, and then insert the state at this position...
	// For each existing tags:
	for (Tag::List::iterator it = Tag::all.begin(); it != Tag::all.end(); ++it) {
		// If the current tag isn't the one to assign or the current one on the item, go to the next tag:
		if (*it != tag && itStates != m_states.end() && *it != (*itStates)->parentTag())
			continue;
		// We found the tag to insert:
		if (*it == tag) {
			// And the item already have the tag:
			if (itStates != m_states.end() && *it == (*itStates)->parentTag()) {
				// We replace the state if wanted:
				if (orReplace) {
					itStates = m_states.insert(itStates, state);
					++itStates;
					m_states.remove(itStates);
					recomputeStyle();
				}
			} else {
				m_states.insert(itStates, state);
				recomputeStyle();
			}
			return;
		}
		// The item has this tag:
		if (itStates != m_states.end() && *it == (*itStates)->parentTag())
			++itStates;
	}
}

QFont Item::font()
{
	return m_computedState.font( basket()->QScrollView::font() );
}

QColor Item::backgroundColor()
{
	if (m_computedState.backgroundColor().isValid())
		return m_computedState.backgroundColor();
	else
		return basket()->backgroundColor();
}

QColor Item::textColor()
{
	if (m_computedState.textColor().isValid())
		return m_computedState.textColor();
	else
		return basket()->textColor();
}

void Item::recomputeStyle()
{
	State::merge(m_states, &m_computedState, &m_emblemsCount, &m_haveInvisibleTags, basket()->backgroundColor());
//	unsetWidth();
	if (content())
		content()->fontChanged();
//	requestRelayout(); // TODO!
}

void Item::recomputeAllStyles()
{
	if (content()) // We do the merge ourself, without calling recomputeStyle(), so there is no infinite recursion:
		State::merge(m_states, &m_computedState, &m_emblemsCount, &m_haveInvisibleTags, basket()->backgroundColor());
	else if (isGroup())
		FOR_EACH_CHILD (child)
			child->recomputeAllStyles();
}


void Item::addTag(Tag *tag)
{
	addState(tag->states().first(), /*but do not replace:*/false);
}

void Item::removeState(State *state)
{
	for (State::List::iterator it = m_states.begin(); it != m_states.end(); ++it)
		if (*it == state) {
			m_states.remove(it);
			recomputeStyle();
			return;
		}
}

void Item::removeTag(Tag *tag)
{
	for (State::List::iterator it = m_states.begin(); it != m_states.end(); ++it)
		if ((*it)->parentTag() == tag) {
			m_states.remove(it);
			recomputeStyle();
			return;
		}
}

void Item::removeAllTags()
{
	m_states.clear();
	recomputeStyle();
}

void Item::addTagToSelectedItems(Tag *tag)
{
	if (content() && isSelected())
		addTag(tag);

	FOR_EACH_CHILD (child)
		child->addTagToSelectedItems(tag);
}

void Item::removeTagFromSelectedItems(Tag *tag)
{
	if (content() && isSelected()) {
		if (hasTag(tag))
			setWidth(0);
		removeTag(tag);
	}

	FOR_EACH_CHILD (child)
		child->removeTagFromSelectedItems(tag);
}

void Item::removeAllTagsFromSelectedItems()
{
	if (content() && isSelected()) {
		if (m_states.count() > 0)
			setWidth(0);
		removeAllTags();
	}

	FOR_EACH_CHILD (child)
		child->removeAllTagsFromSelectedItems();
}

void Item::addStateToSelectedItems(State *state, bool orReplace)
{
	if (content() && isSelected())
		addState(state, orReplace);

	FOR_EACH_CHILD (child)
		child->addStateToSelectedItems(state, orReplace); // TODO: Basket::addStateToSelectedItems() does not have orReplace
}

void Item::changeStateOfSelectedItems(State *state)
{
	if (content() && isSelected() && hasTag(state->parentTag()))
		addState(state);

	FOR_EACH_CHILD (child)
		child->changeStateOfSelectedItems(state);
}

bool Item::selectedItemsHaveTags()
{
	if (content() && isSelected() && m_states.count() > 0)
		return true;

	FOR_EACH_CHILD (child)
		if (child->selectedItemsHaveTags())
			return true;
	return false;
}



bool Item::hasState(State *state)
{
	for (State::List::iterator it = m_states.begin(); it != m_states.end(); ++it)
		if (*it == state)
			return true;
	return false;
}

bool Item::hasTag(Tag *tag)
{
	for (State::List::iterator it = m_states.begin(); it != m_states.end(); ++it)
		if ((*it)->parentTag() == tag)
			return true;
	return false;
}

State* Item::stateOfTag(Tag *tag)
{
	for (State::List::iterator it = m_states.begin(); it != m_states.end(); ++it)
		if ((*it)->parentTag() == tag)
			return *it;
	return 0;
}

State* Item::stateForEmblemNumber(int number)
{
	int i = -1;
	for (State::List::Iterator it = m_states.begin(); it != m_states.end(); ++it)
		if (!(*it)->emblem().isEmpty()) {
			++i;
			if (i == number)
				return *it;
		}
	return 0;
}

void Item::inheritTagsOf(Item *item)
{
	if (!item || !content())
		return;

	for (State::List::iterator it = item->states().begin(); it != item->states().end(); ++it)
		if ((*it)->parentTag() && (*it)->parentTag()->inheritedBySiblings())
			addTag((*it)->parentTag());
}

void Item::unbufferizeAll()
{
	unbufferize();

	if (isGroup()) {
		Item *child = firstChild();
		while (child) {
			child->unbufferizeAll();
			child = child->next();
		}
	}
}

void Item::bufferizeSelectionPixmap()
{
	if (m_bufferedSelectionPixmap.isNull()) {
		QColor insideColor = KGlobalSettings::highlightColor();
		KPixmap kpixmap(m_bufferedPixmap);
		m_bufferedSelectionPixmap = KPixmapEffect::fade(kpixmap, 0.25, insideColor);
	}
}

void Item::recomputeBlankRects(QValueList<QRect> &blankAreas)
{
	if (!matching())
		return;

	substractRectOnAreas(rect(), blankAreas);
	if (hasResizer())
		substractRectOnAreas(resizerRect(), blankAreas);

	if (isGroup()) {
		Item *child = firstChild();
		bool first = true;
		while (child) {
			if ((showSubItems() || first) && child->matching())
				child->recomputeBlankRects(blankAreas);
			child = child->next();
			first = false;
		}
	}
}

void Item::linkLookChanged()
{
	if (isGroup()) {
		Item *child = firstChild();
		while (child) {
			child->linkLookChanged();
			child = child->next();
		}
	} else
		content()->linkLookChanged();
}

Item* Item::itemForFullPath(const QString &path)
{
	if (content() && fullPath() == path)
		return this;

	Item *child = firstChild();
	Item *found;
	while (child) {
		found = child->itemForFullPath(path);
		if (found)
			return found;
		child = child->next();
	}
	return 0;
}

void Item::exportToHTML(QTextStream &stream, int indent, const HtmlExportData &exportData)
{
	QString spaces;

	if (isColumn()) {
		QString width = "";
		if (false/*TODO: DEBUG AND REENABLE: hasResizer()*/) {
			// As we cannot be precise in CSS (say eg. "width: 50%-40px;"),
			// we output a percentage that is approximatively correct.
			// For instance, we compute the currently used percentage of width in the basket
			// and try make make it the same on a 1024*768 display in a Web browser:
			int availableSpaceForColumnsInThisBasket = basket()->contentsWidth() - (basket()->columnsCount() - 1) * RESIZER_WIDTH;
			int availableSpaceForColumnsInBrowser    = 1024    /* typical screen width */
			                                           - 25    /* window border and scrollbar width */
			                                           - 2 * 5 /* page margin */
			                                           - (basket()->columnsCount() - 1) * RESIZER_WIDTH;
			if (availableSpaceForColumnsInThisBasket <= 0)
				availableSpaceForColumnsInThisBasket = 1;
			int widthValue = (int)(availableSpaceForColumnsInBrowser * (double) groupWidth() / availableSpaceForColumnsInThisBasket);
			if (widthValue <= 0)
				widthValue = 1;
			if (widthValue > 100)
				widthValue = 100;
			width = QString(" width=\"%1%\"").arg(QString::number(widthValue));
		}
		stream << spaces.fill(' ', indent) << "<td class=\"column\"" << width << ">\n";

		FOR_EACH_CHILD (child) {
			stream << spaces.fill(' ', indent + 1);
			child->exportToHTML(stream, indent + 1, exportData);
			stream << '\n';
		}

		stream << spaces.fill(' ', indent) << "</td>\n";
		if (hasResizer())
			stream << spaces.fill(' ', indent) << "<td class=\"columnHandle\"></td>\n";
		return;
	}

	QString freeStyle;
	if (isFree())
		freeStyle = " style=\"position: absolute; left: " + QString::number(x()) + "px; top: " + QString::number(y()) + "px; width: " + QString::number(groupWidth()) + "px\"";

	if (isGroup()) {
		stream << '\n' << spaces.fill(' ', indent) << "<table" << freeStyle << ">\n"; // Item content is expected to be on the same HTML line, but NOT groups
		int i = 0;
		FOR_EACH_CHILD (child) {
			stream << spaces.fill(' ', indent);
			if (i == 0)
				stream << " <tr><td class=\"groupHandle\"><img src=\"" << exportData.imagesFolderName << (isFolded() ? "expand_group.png" : "fold_group.png")
				       << "\" width=\"" << EXPANDER_WIDTH << "\" height=\"" << EXPANDER_HEIGHT << "\"></td>\n";
			else if (i == 1)
				stream << " <tr><td class=\"freeSpace\" rowspan=\"" << countDirectChilds() << "\"></td>\n";
			else
				stream << " <tr>\n";
			stream << spaces.fill(' ', indent) << "  <td>";
			child->exportToHTML(stream, indent + 3, exportData);
			stream << "</td>\n"
			       << spaces.fill(' ', indent) << " </tr>\n";
			++i;
		}
		stream << '\n' << spaces.fill(' ', indent) << "</table>\n" /*<< spaces.fill(' ', indent - 1)*/;
	} else {
		// Additionnal class for the content (link, netword, color...):
		QString additionnalClasses = content()->cssClass();
		if (!additionnalClasses.isEmpty())
			additionnalClasses = " " + additionnalClasses;
		// Assign the style of each associted tags:
		for (State::List::Iterator it = m_states.begin(); it != m_states.end(); ++it)
			additionnalClasses += " tag_" + (*it)->id();
		//stream << spaces.fill(' ', indent);
		stream << "<table class=\"item" << additionnalClasses << "\"" << freeStyle << "><tr>";
		if (m_emblemsCount > 0) {
			stream << "<td class=\"tags\"><nobr>";
			for (State::List::Iterator it = m_states.begin(); it != m_states.end(); ++it)
				if (!(*it)->emblem().isEmpty()) {
					int emblemSize = 16;
					QString iconFileName = Basket::copyIcon((*it)->emblem(), emblemSize, exportData.iconsFolderPath);
					stream << "<img src=\"" << exportData.iconsFolderName << iconFileName
					       << "\" width=\"" << emblemSize << "\" height=\"" << emblemSize
					       << "\" alt=\"" << (*it)->textEquivalent() << "\" title=\"" << (*it)->fullName() << "\">";
				}
				stream << "</nobr></td>";
		}
		stream << "<td>";
		content()->exportToHTML(stream, indent, exportData);
		stream << "</td></tr></table>";
		/*if (m_emblemsCount > 0) {
			stream << "<table class=\"item" << additionnalClasses << "\"" << freeStyle << "><tr><td class=\"tags\">";
			for (State::List::Iterator it = m_states.begin(); it != m_states.end(); ++it)
				if (!(*it)->emblem().isEmpty()) {
					int emblemSize = 16;
					QString iconFileName = Basket::copyIcon((*it)->emblem(), emblemSize, exportData.iconsFolderPath);
					stream << "<img src=\"" << exportData.iconsFolderName << iconFileName
					       << "\" width=\"" << emblemSize << "\" height=\"" << emblemSize
					       << "\" alt=\"" << (*it)->textEquivalent() << "\" title=\"" << (*it)->fullName() << "\">";
				}
			stream << "</td><td>";
			content()->exportToHTML(stream, indent, exportData);
			stream << "</td></tr></table>";
		} else {
			stream << "<div class=\"item" << additionnalClasses << "\"" << freeStyle << ">";
			content()->exportToHTML(stream, indent, exportData);
			stream << "</div>";
		}*/
	}
}

void Item::usedStates(QValueList<State*> &states)
{
	if (content())
		for (State::List::Iterator it = m_states.begin(); it != m_states.end(); ++it)
			if (!states.contains(*it))
				states.append(*it);

	FOR_EACH_CHILD (child)
		child->usedStates(states);
}







#if 0

/** Item */

QString Item::toHtml(const QString &imageName)
{
	switch (m_type) {
		case Text:
			/*return "<font color=" + backgroundColor().name() + + font().name() + font + ">" +
			Utilities::textToHTMLWithoutP(text()) + "</font>";*/
			return Utilities::textToHTMLWithoutP(text());
		case Html:
			return Utilities::htmlToParagraph(html());
		case Image:
		case Animation:
			{
				if ( (m_type == Image     && pixmap() == 0L) ||
				     (m_type == Animation && movie()  == 0L)    ) {
					QMimeSourceFactory::defaultFactory()->setData(imageName, 0L);
					return i18n("(Image)"); // Image or animation not yet loaded!!
				}

				QImage image;
				if (m_type == Image)
					image = pixmap()->convertToImage();
				else
					image = movie()->framePixmap().convertToImage();
				image = image.smoothScale(200, 150, QImage::ScaleMin);
				QPixmap pixmap = QPixmap(image);
				QMimeSourceFactory::defaultFactory()->setPixmap(imageName, pixmap);
				return "<img src=" + imageName + ">"; ///

/*				// FIXME: movie isn't loaded yet: CRASH!
				return i18n("(Image)");
				// Not executed, because don't work:
				QImage image;
				if (m_type == Image)
					image = pixmap()->convertToImage();
				else
					image = movie()->framePixmap().convertToImage();
				image = image.smoothScale(200, 150, QImage::ScaleMin);
				QPixmap pixmap = QPixmap(image);
				QMimeSourceFactory::defaultFactory()->setPixmap(imageName, pixmap);
				return "<img src=" + imageName + ">"; ///
	*/			//TODO?: QMimeSourceFactory::defaultFactory()->setData(imageName, 0L);
			}
		case Sound:
		case File:
			{
				/// FIXME: Since fullPath() doesn't exist yet, the icon rely on the extension.
				///        Bad if there isn't one or if it's a wrong one.
				/*QPixmap icon = DesktopIcon(
					ItemFactory::iconForURL(fullPath()),
					(m_type == Sound ? LinkLook::soundLook : LinkLook::fileLook)->iconSize());
				QMimeSourceFactory::defaultFactory()->setPixmap(imageName, icon);
				return "<img src=" + imageName + "> " + fileName(); */ ///
				return m_linkLabel->toHtml(imageName);
			}
		case Link:
			{
				QString link = m_linkLabel->toHtml(imageName);
				if (!autoTitle() && title() != ItemFactory::titleForURL(url().prettyURL()))
					link += "<br><i>" + url().prettyURL() + "</i>"; ///
				return link;
			}
		case Launcher:
			{
				return m_linkLabel->toHtml(imageName);
				//KService service(fullPath()); // service.icon()
				//return service.name() + "<br><i>" + service.exec() + "</i>"; ///
			}
		case Color:
			return "<b><font color=" + color().name() + ">" + color().name() + "</font></b>";
		case Unknown:
			return text();
	}
	return QString();
}

void Item::setFocused(bool focused)
{
	if (focused == m_isFocused)
		return;

	m_isFocused = focused;
	update();
}

void Item::delayedEnsureVisible()
{
	m_parentBasket->ensureVisibleItem(this);
}

#endif // #if 0

//# include " item.moc "
